#include <cassert>
#include <cstring>
#include <cstdlib>
#include <cstdio>
#include "xiobj.h"

using namespace std;

struct XiOutputStream;
struct XiInputStream;
struct XiFileInput;
struct XiFileOutput;

// These decls to keep bindgen happy
struct OutputStream;
struct InputStream;
struct FileInput;
struct FileOutput;


typedef int64_t *ixstring;

// Maybe should have just wrapped FILE * instead of creating
// these classes to wrap.  Would have avoided a bunch of crap.
// Problem is that makebindings seems to want there to be an actual class
// named OutputStream.

struct XiOutputStream_vtable;
struct XiOutputStream: public XiObj
{
    XiOutputStream_vtable* vtable();
    XiOutputStream();
    XiOutputStream(FILE *f);
    OutputStream* impl();
    static void invokeDtor(void* mePtr, void*);
    static void __classOp(ClassOp op, void* memory);
    typedef OutputStream wrappedType;
    FILE *file;
    OutputStream *getImpl() { return 0;}
    void createImpl() { assert(false); }
    static void puts(XiOutputStream* __thisPtr, void* s);
    static void putc(XiOutputStream* __thisPtr, int64_t c);
    static void flush(XiOutputStream* __thisPtr);
};

inline OutputStream* fromXI(XiOutputStream* wrapped) {
    return wrapped ? wrapped->impl() : 0;
}

struct XiOutputStream_vtable: public XiVTable
{
    void (*__classOp)(ClassOp, void*);
    void (*puts)(XiOutputStream*, void*);
    void (*putc)(XiOutputStream*, int64_t);
    void (*flush)(XiOutputStream*);
};

inline XiOutputStream_vtable* XiOutputStream::vtable() {
    return static_cast<XiOutputStream_vtable*>(vptr);
}

extern XiOutputStream_vtable XI(_vt_OutputStream);
extern int64_t XI(_size_OutputStream);
XI_EXPORT void XI(_init_OutputStream)() __attribute__((constructor));


static int64_t endsLine(int64_t c) { // including EOF
	return c == '\n' || c == '\r' || c == EOF;
}

ixstring str_alloc(int64_t n) {
	int64_t *base = (int64_t *)XI(_alloc_i)((n+1) * sizeof(int64_t));
	base[0] = n;
	return &base[1];
}

ixstring str_realloc(ixstring s, int64_t n) {
	int64_t *newb = str_alloc(n+1);
	int64_t *ret = &newb[1];
	size_t bytes = s[-1] * sizeof(int64_t);
	if (bytes > n * sizeof(int64_t)) bytes = n * sizeof(int64_t);
	memcpy(ret, s, bytes);
	return ret;
}

// Converts an Xi/Xi string to a C string.
// Requires: s is a string containing only ASCII characters.
char *str_toC(ixstring s) {
	char *ret = (char *)XI(_alloc_i)(s[-1] + 1);
	for (int64_t i = 0; i < s[-1]; i++) {
		ret[i] = (char)s[i];
	}
	ret[s[-1] ] = 0;
	return ret;
}

struct XiInputStream_vtable;
struct XiInputStream: public XiObj
{
    XiInputStream_vtable* vtable();
    XiInputStream();
    XiInputStream(FILE *f);
    InputStream* impl();
    static void invokeDtor(void* mePtr, void*);
    static void __classOp(ClassOp op, void* memory);
    typedef InputStream wrappedType;
    FILE *file;
    InputStream *getImpl() { return 0;}
    void createImpl() { assert(false); }

    // Since feof seems to only detect EOF after it happened,
    // we make feof read ahead; and so we potentially buffer a character.
    bool hasBufferedChar; // inited to false by GC
    int64_t bufferedChar;

    int64_t bufGetC();
    static void* gets(XiInputStream* __thisPtr);
    static int64_t getc(XiInputStream* __thisPtr);
    static bool feof(XiInputStream* __thisPtr);
};

inline InputStream* fromXI(XiInputStream* wrapped) {
    return wrapped ? wrapped->impl() : 0;
}

struct XiInputStream_vtable: public XiVTable
{
    void (*__classOp)(ClassOp, void*);
    void* (*gets)(XiInputStream*);
    int64_t (*getc)(XiInputStream*);
    bool (*feof)(XiInputStream*);
};

inline XiInputStream_vtable* XiInputStream::vtable() {
    return static_cast<XiInputStream_vtable*>(vptr);
}

extern XiInputStream_vtable XI(_vt_InputStream);
extern int64_t XI(_size_InputStream);
XI_EXPORT void XI(_init_InputStream)() __attribute__((constructor));




struct XiFileInput_vtable;
struct XiFileInput: public XiInputStream
{
    XiFileInput_vtable* vtable();
    XiFileInput();
    FileInput* impl();
    static void invokeDtor(void* mePtr, void*);
    static void __classOp(ClassOp op, void* memory);
    typedef FileInput wrappedType;
    FileInput *getImpl() { return 0; }
    static void close(XiFileInput* __thisPtr);
};

inline FileInput* fromXI(XiFileInput* wrapped) {
    return wrapped ? wrapped->impl() : 0;
}

struct XiFileInput_vtable: public XiInputStream_vtable
{
    void* reserved;
    void (*close)(XiFileInput*);
};

inline XiFileInput_vtable* XiFileInput::vtable() {
    return static_cast<XiFileInput_vtable*>(vptr);
}

extern XiFileInput_vtable XI(_vt_FileInput);
extern int64_t XI(_size_FileInput);
XI_EXPORT void XI(_init_FileInput)() __attribute__((constructor));




struct XiFileOutput_vtable;
struct XiFileOutput: public XiOutputStream
{
    XiFileOutput_vtable* vtable();
    XiFileOutput();
    FileOutput* impl();
    static void invokeDtor(void* mePtr, void*);
    static void __classOp(ClassOp op, void* memory);
    typedef FileOutput wrappedType;
    FileOutput *getImpl() { return (FileOutput *)this; }
    static void close(XiFileOutput* __thisPtr);
};

inline FileOutput* fromXI(XiFileOutput* wrapped) {
    return wrapped ? wrapped->impl() : 0;
}

struct XiFileOutput_vtable: public XiOutputStream_vtable
{
    void* reserved;
    void (*close)(XiFileOutput*);
};

inline XiFileOutput_vtable* XiFileOutput::vtable() {
    return static_cast<XiFileOutput_vtable*>(vptr);
}

extern XiFileOutput_vtable XI(_vt_FileOutput);
extern int64_t XI(_size_FileOutput);
XI_EXPORT void XI(_init_FileOutput)() __attribute__((constructor));


XI_EXPORT XiFileInput* XI(readFile_o9FileInputai)(void* filename);
XI_EXPORT XiFileOutput* XI(createFile_o10FileOutputai)(void* filename);
XI_EXPORT XiFileOutput* XI(appendFile_o10FileOutputai)(void* filename);
XI_EXPORT bool XI(removeFile_bai)(void* filename);

XI_EXPORT XiOutputStream *XI(stdout_o12OutputStream)();
XI_EXPORT XiInputStream *XI(stdin_o11InputStream)();
